// Oleg Petruny proprietary.


#include "InteractableScreenCapturer.h"

#include "Engine/Texture.h"
#include "Engine/TextureRenderTarget2D.h"
#include "Kismet/GameplayStatics.h"
#include "opencv2/imgproc.hpp"
#include "RHIResources.h"
#include "SceneView.h"
#include "TextureCompressorModule.h"

#include "Interactable.h"

constexpr float tickInterval = 1.0f / 5;
constexpr float textureWidth = 1280 / 2;
constexpr float textureHeight = 720 / 2;
constexpr float rayDistance = 7000;

UInteractableScreenCapturer::UInteractableScreenCapturer(const FObjectInitializer& ObjectInitializer)
    : USceneCaptureComponent2D(ObjectInitializer)
{
    PrimaryComponentTick.bCanEverTick = false;
    PrimaryComponentTick.bStartWithTickEnabled = false;

    TextureTarget = LoadObject<UTextureRenderTarget2D>(this, TEXT("/Game/tempStaff/Input"));
    //TextureTarget = ObjectInitializer.CreateDefaultSubobject<UTextureRenderTarget2D>(this, TEXT("UInteractableScreenCapturer_TextureRenderTarget"));
    //TextureTarget->RenderTargetFormat = ETextureRenderTargetFormat::RTF_RGBA8;
    //TextureTarget->ClearColor = FLinearColor::Black;
    //TextureTarget->bAutoGenerateMips = false;
    //TextureTarget->bCanCreateUAV = false;
    //TextureTarget->InitAutoFormat(textureWidth, textureHeight);
    //TextureTarget->UpdateResourceImmediate(true);

    PrimitiveRenderMode = ESceneCapturePrimitiveRenderMode::PRM_LegacySceneCapture;
    bCaptureEveryFrame = 0;
    bCaptureOnMovement = 0;
    LODDistanceFactor = 10000;
    bUseRayTracingIfEnabled = false;
    CaptureSource = ESceneCaptureSource::SCS_BaseColor;

    CompositeMode = ESceneCaptureCompositeMode::SCCM_Overwrite;
    PostProcessBlendWeight = 0;
    bUseCustomProjectionMatrix = false;
    bAlwaysPersistRenderingState = true;

    ShowFlags = { EShowFlagInitMode::ESFIM_All0 };
    ShowFlags.SetBSP(true);
    ShowFlags.SetLandscape(true);
    ShowFlags.SetSkeletalMeshes(true);
    ShowFlags.SetStaticMeshes(true);
    ShowFlags.SetInstancedFoliage(true);
    ShowFlags.SetInstancedGrass(true);
    ShowFlags.SetInstancedStaticMeshes(true);
    ShowFlags.SetPaper2DSprites(true);
    ShowFlags.SetGame(true);
    ShowFlags.SetMaterials(true);
    ShowFlags.SetRendering(true);

    PrimaryComponentTick.TickInterval = tickInterval;
    PrimaryComponentTick.bCanEverTick = true;
    PrimaryComponentTick.bStartWithTickEnabled = true;
}

void UInteractableScreenCapturer::TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
    switch(TickType)
    {
        case ELevelTick::LEVELTICK_TimeOnly:
        case ELevelTick::LEVELTICK_ViewportsOnly:
        case ELevelTick::LEVELTICK_All:
        {
            USceneCaptureComponent2D::TickComponent(DeltaTime, TickType, ThisTickFunction);
            CaptureScene();
            GetCameraView(DeltaTime, _view);
            Process();
        }
        break;
        default:
            break;
    }
}

void UInteractableScreenCapturer::Process()
{
    static FThreadSafeBool running = false;
    if(!running)
    {
        running = true;
        ENQUEUE_RENDER_COMMAND(GetInteractablesFromScreen)(
            [
                capture = TextureTarget->GetResource()->TextureRHI,
                    world = GetWorld(),
                    view = _view,
                    this
            ]
            (FRHICommandListImmediate& RHICmdList)
            {
                double StartTime = FPlatformTime::Seconds();
                static FIntPoint _extent = capture->GetDesc().Extent;
                static uint32 _size = _extent.X * _extent.Y;

                static uint8* _bitMap = new uint8[_size]; // cv::Mat allocation is weird sometimes...

                uint32 stride = 0;
                FColor* data = (FColor*)RHICmdList.LockTexture2D(
                    capture,
                    0,
                    RLM_ReadOnly,
                    stride,
                    false);
                for(uint32 i = 0; i < _size; ++i)
                {
                    switch(data[i].R)
                    {
                        case 0xFF:
                            switch(data[i].G)
                            {
                                case 0x00:
                                    switch(data[i].B)
                                    {
                                        case 0xFF:
                                            _bitMap[i] = static_cast<uint8>(0xFF);
                                            continue;
                                        default:
                                            break;
                                    }
                                    [[fallthrough]];
                                default:
                                    break;
                            }
                            [[fallthrough]];
                        default:
                            _bitMap[i] = static_cast<uint8>(0x00);
                            break;
                    }
                }
                RHICmdList.UnlockTexture2D(capture, 0, false);

                cv::Mat componentsMap, componentsStats, componentsCentroids;
                static cv::Mat bitMapCVProxy{ _extent.Y, _extent.X, CV_8UC1, _bitMap };
                auto components = cv::connectedComponentsWithStats(
                    bitMapCVProxy,
                    componentsMap,
                    componentsStats,
                    componentsCentroids,
                    4,
                    CV_16U
                );

                //static FVector _oldLocation;
                //static FRotator _oldRotation;
                //static FMatrix _invProjection;
                //if(view.Location != _oldLocation || view.Rotation != _oldRotation)
                //{
                //    _oldLocation = view.Location;
                //    _oldRotation = view.Rotation;
                //    _invProjection =
                //        (FTranslationMatrix(-(view.Location))
                //            * (FInverseRotationMatrix(view.Rotation) * FMatrix(FPlane(0, 0, 1, 0), FPlane(1, 0, 0, 0), FPlane(0, 1, 0, 0), FPlane(0, 0, 0, 1)))
                //            * view.CalculateProjectionMatrix())
                //        .InverseFast();
                //}
                FMatrix _invProjection =
                    (FTranslationMatrix(-(view.Location))
                        * (FInverseRotationMatrix(view.Rotation) * FMatrix(FPlane(0, 0, 1, 0), FPlane(1, 0, 0, 0), FPlane(0, 1, 0, 0), FPlane(0, 0, 0, 1)))
                        * view.CalculateProjectionMatrix())
                    .InverseFast();
                GEngine->AddOnScreenDebugMessage(60, 0.5f, FColor::Yellow, FString::Printf(TEXT("x:%f, y:%f"), view.Rotation.Yaw, view.Rotation.Roll));
                FVector worldLoc, worldDir;
                for(int i = 1; i < components; ++i)
                {
                    FVector2D screenPos = { ((double*)componentsCentroids.data)[i * 2], ((double*)componentsCentroids.data)[i * 2 + 1] };
                    FSceneView::DeprojectScreenToWorld(screenPos, FIntRect{ {0, 0}, _extent }, _invProjection, worldLoc, worldDir);
                    //GEngine->AddOnScreenDebugMessage(66 + i, 0.5f, FColor::Yellow, FString::Printf(TEXT("x:%f, y:%f"), screenPos.X, screenPos.Y));

                    FHitResult result{};
                    FVector startLocation = view.Location;
                    FVector endLocation = worldLoc + worldDir * rayDistance;
                    world->LineTraceSingleByChannel(
                        result,
                        startLocation,
                        endLocation,
                        ECC_GameTraceChannel1
                    );
                    if(result.bBlockingHit)
                    {
                        if(auto interactable = Cast<AInteractable>(result.GetActor()))
                        {
                            if(!_sawInteractables.Contains(interactable))
                            {
                                _sawInteractables.Add(interactable);
                                AsyncTask(ENamedThreads::GameThread, [=, this]()
                                    {
                                        interactableInScreenDelegate.Execute(interactable);
                                    });
                            }
                            //GEngine->AddOnScreenDebugMessage(10 + i, 0.5f, FColor::Yellow, interactable->GetName());
                            GEngine->AddOnScreenDebugMessage(20 + i, 0.5f, FColor::Yellow, FString::Printf(TEXT("dist:%f"), (view.Location - interactable->GetActorLocation()).Length()));
                        }

                        DrawDebugLine(world, startLocation, endLocation, FColor::Green, false, tickInterval, 0, 0.5f);
                    }
                    else
                    {
                        DrawDebugLine(world, startLocation, endLocation, FColor::Red, false, tickInterval, 10, 0.5f);
                    }
                    //DrawDebugLine(world, startLocation, endLocation, FColor::Red, false, tickInterval, 0, 0.5f);
                }
                GEngine->AddOnScreenDebugMessage(667, 0.5f, FColor::Yellow, FString::Printf(TEXT("time:%f"), FPlatformTime::Seconds() - StartTime));
                running = false;
            }
        );
    }
}